import { Plugin, PluginKey } from 'prosemirror-state';
import { isEqual } from 'lodash';
import { Decoration, DecorationSet } from 'prosemirror-view';
import { findChildren, findParentNodeClosestToPos } from 'prosemirror-utils';
import { cn } from '../../../classname';
import { isNodeEmpty } from '../../../utils/nodes';
import { isTextSelection } from '../../../utils/selection';
import { getPlaceholderContent } from '../../../utils/placeholder';
import './index.css';
const getPlaceholderPluginKeys = (schema) => {
    var _a;
    const pluginKeys = [];
    for (const node in schema.nodes) {
        if (schema.nodes[node]) {
            const spec = schema.nodes[node].spec;
            if ((_a = spec.placeholder) === null || _a === void 0 ? void 0 : _a.customPlugin) {
                pluginKeys.push(spec.placeholder.customPlugin);
            }
        }
    }
    return pluginKeys;
};
const b = cn('placeholder');
export const createPlaceholder = (node, parent, focus) => {
    const content = getPlaceholderContent(node, parent);
    if (!content)
        return null;
    const placeholder = document.createElement('div');
    placeholder.classList.add(...b({ focus }).split(' '));
    placeholder.textContent = content;
    return placeholder;
};
const placeholderNeeded = (node) => {
    const childrenWithPlaceholderVisible = findChildren(node, (n) => { var _a; return Boolean((_a = n.type.spec.placeholder) === null || _a === void 0 ? void 0 : _a.alwaysVisible); });
    return (isNodeEmpty(node) &&
        // Если есть дочерние ноды у которых есть постоянный плейсхолдер - отдаем им приоритет
        !childrenWithPlaceholderVisible.length);
};
const addDecoration = (widgetsMap, node, pos, parent, cursorPos, globalState) => {
    const placeholderSpec = node.type.spec.placeholder;
    const decorationPosition = pos + node.childCount + 1;
    // Не добавляем декорацию если на этой позиции уже есть плейсхолдер
    if (!placeholderSpec || widgetsMap[decorationPosition])
        return;
    if (placeholderSpec.customPlugin) {
        widgetsMap[decorationPosition] = placeholderSpec.customPlugin;
        return;
    }
    const focus = decorationPosition === cursorPos;
    const placeholderDOM = createPlaceholder(node, parent, focus);
    if (!placeholderDOM)
        return;
    if (focus)
        globalState.hasFocus = true;
    widgetsMap[decorationPosition] = {
        pos: decorationPosition,
        toDOM: placeholderDOM,
        spec: { focus },
    };
};
const pluginKey = new PluginKey('placeholder_plugin');
export const Placeholder = (builder) => {
    builder.addPlugin(() => new Plugin({
        key: pluginKey,
        props: {
            attributes(state) {
                const attrs = {};
                if (pluginKey.getState(state).hasFocus) {
                    // hide native cursor
                    attrs.class = 'yfm-editor-hidecursor';
                }
                return attrs;
            },
            decorations(state) {
                var _a;
                return (_a = pluginKey.getState(state)) === null || _a === void 0 ? void 0 : _a.decorationSet;
            },
        },
        state: {
            init: (_config, state) => initState(state),
            apply: applyState,
        },
    }));
};
function getPlaceholderWidgetSpecs(state) {
    var _a;
    const globalState = { hasFocus: false };
    const widgetsMap = {};
    const { selection } = state;
    const cursorPos = isTextSelection(selection) ? (_a = selection.$cursor) === null || _a === void 0 ? void 0 : _a.pos : null;
    getPlaceholderPluginKeys(state.schema).forEach((f) => {
        var _a;
        // Используем find потому что при помощи него можно проитерировать по DecorationSet.
        (_a = f.getState(state)) === null || _a === void 0 ? void 0 : _a.find(undefined, undefined, (spec) => {
            widgetsMap[spec.pos] = f;
            return false;
        });
    });
    // Рисуем плэйсхолдеры для всех нод у которых плэйсхолдер alwaysVisible
    const decorate = (node, pos, parent) => {
        const placeholderSpec = node.type.spec.placeholder;
        if (placeholderSpec && placeholderSpec.alwaysVisible && placeholderNeeded(node)) {
            addDecoration(widgetsMap, node, pos, parent, cursorPos, globalState);
        }
    };
    state.doc.descendants(decorate);
    const parentNode = findParentNodeClosestToPos(state.selection.$from, (node) => {
        return Boolean(node.type.spec.placeholder);
    });
    const placeholderSpec = parentNode === null || parentNode === void 0 ? void 0 : parentNode.node.type.spec.placeholder;
    // Дорисовываем плэйсхолдер, если его нужно отрисовать на месте курсора
    if (parentNode &&
        placeholderNeeded(parentNode.node) &&
        placeholderSpec &&
        !placeholderSpec.alwaysVisible) {
        const { node, pos, depth } = parentNode;
        const parent = depth > 0 ? state.selection.$from.node(depth - 1) : null;
        addDecoration(widgetsMap, node, pos, parent, cursorPos, globalState);
    }
    const widgetSpecs = Object.values(widgetsMap).filter((decoration) => !(decoration instanceof PluginKey));
    return { widgetSpecs, hasFocus: globalState.hasFocus };
}
function initState(state) {
    const { widgetSpecs, hasFocus } = getPlaceholderWidgetSpecs(state);
    const decorationSet = DecorationSet.create(state.doc, widgetSpecs.map((widget) => Decoration.widget(widget.pos, widget.toDOM, widget.spec)));
    return { decorationSet, hasFocus };
}
function applyState(tr, oldPluginState, _oldState, newState) {
    const { widgetSpecs, hasFocus } = getPlaceholderWidgetSpecs(newState);
    const { decorationSet } = oldPluginState;
    const oldMappedSet = decorationSet.map(tr.mapping, tr.doc);
    // Find all decorations that are present in old and new set
    const decorationsThatDidNotChange = widgetSpecs.reduce((a, { pos, spec }) => {
        const deco = oldMappedSet.find(pos, pos);
        if (deco.length && isEqual(deco[0].spec, spec))
            a.push(...deco);
        return a;
    }, []);
    // Those are decorations that are presenr only in new set
    const newAddedDecorations = widgetSpecs.filter(({ pos }) => !decorationsThatDidNotChange.map(({ from }) => from).includes(pos));
    // That is a set with decorations that are present in old set and absent in new set
    const notRelevantDecorations = oldMappedSet.remove(decorationsThatDidNotChange);
    let newSet = oldMappedSet;
    // Remove decorations that are not present in new set
    if (notRelevantDecorations.find().length)
        newSet = newSet.remove(notRelevantDecorations.find());
    // Add new decorations
    if (newAddedDecorations.length)
        newSet = newSet.add(tr.doc, newAddedDecorations.map((widget) => Decoration.widget(widget.pos, widget.toDOM, widget.spec)));
    return { decorationSet: newSet, hasFocus };
}
